from os import environ
from typing import TypeVar, Callable

DT = TypeVar("DT")
RT = TypeVar("RT")
Parser = Callable[[str], RT]


class Undefined:
    pass


UNDEFINED = Undefined()


def get_var(
    key: str,
    default: DT = UNDEFINED,
    parser: Parser[RT] = lambda x: x,
) -> DT | RT:
    if key not in environ:
        if isinstance(default, Undefined):
            raise EnvError(key)
        return default

    value = environ[key]

    try:
        return parser(value)
    except ValueError as e:
        raise ParseError(key, e)


class EnvError(KeyError):
    def __init__(self, key: str):
        self.key = key

    def __str__(self):
        return f"Environment variable `{self.key}` not set"


class ParseError(ValueError):
    def __init__(self, key: str, cause: ValueError):
        self.key = key
        self.cause = cause

    def __str__(self):
        return f"Can't parse environment variable `{self.key}`: {self.cause}"



import env

assert env.get("TOKEN") == "abc123"
assert env.get("X", False) is False
assert env.get_int("PORT") == 123
assert env.get_strs("KEYS", []) == ["word1", "F402", "12"]
assert env.get_ints("ADMIN_IDS") == [1, 2, 3]
assert env.get_ints("X", 0) == 0

try:
    env.get("X")
except KeyError as e:
    assert str(e) == "Environment variable `X` not set"

try:
    env.get_ints("BAD_NUMS")
except ValueError as e:
    assert (
        str(e) == "Can't parse environment variable `BAD_NUMS`: "
        "invalid literal for int() with base 10: 'a'"
    )



from .core import get_var, DT, UNDEFINED
from .parsers import parse_ints, parse_strs


def get(key: str, default: DT = UNDEFINED) -> str | DT:
    return get_var(key, default)


def get_int(key: str, default: DT = UNDEFINED) -> int | DT:
    return get_var(key, default, int)


def get_strs(key: str, default: DT = UNDEFINED) -> list[str] | DT:
    return get_var(key, default, parse_strs)


def get_ints(key: str, default: DT = UNDEFINED) -> list[int] | DT:
    return get_var(key, default, parse_ints)
